"""ROS2 tf2_msgs message definitions.

Auto-generated by scripts/generate_messages.py
Source: https://github.com/ros2/geometry2/tree/galactic/tf2_msgs
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Literal

import pybag.ros2.galactic.geometry_msgs as geometry_msgs
import pybag.types as t


@dataclass(kw_only=True)
class TF2Error:
    __msg_name__ = 'tf2_msgs/msg/TF2Error'

    NO_ERROR: t.Constant[t.uint8] = 0
    LOOKUP_ERROR: t.Constant[t.uint8] = 1
    CONNECTIVITY_ERROR: t.Constant[t.uint8] = 2
    EXTRAPOLATION_ERROR: t.Constant[t.uint8] = 3
    INVALID_ARGUMENT_ERROR: t.Constant[t.uint8] = 4
    TIMEOUT_ERROR: t.Constant[t.uint8] = 5
    TRANSFORM_ERROR: t.Constant[t.uint8] = 6
    error: t.uint8
    error_string: t.string


@dataclass(kw_only=True)
class TFMessage:
    __msg_name__ = 'tf2_msgs/msg/TFMessage'

    transforms: t.Array[geometry_msgs.TransformStamped]
