"""ROS2 builtin_interfaces message definitions.

Auto-generated by scripts/generate_messages.py
Source: https://github.com/ros2/rcl_interfaces/tree/galactic/builtin_interfaces
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Literal

import pybag.types as t


@dataclass(kw_only=True)
class Duration:
    __msg_name__ = 'builtin_interfaces/msg/Duration'

    sec: t.int32
    nanosec: t.uint32


@dataclass(kw_only=True)
class Time:
    __msg_name__ = 'builtin_interfaces/msg/Time'

    sec: t.int32
    nanosec: t.uint32
